/*
   Copyright (c) 2014 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "includes.h"

#include "svpwm.h"
#include "pidq.h"
#include "vector.h"
#include "fp_math.h"

extern void DSP28x_usDelay(unsigned long aCount);

#define INV_SQRT3_Q15 (int16_t)(1.0 / 1.73 * 32768.0)

PLX_SVPWM_Obj_t SvpwmObj;
PLX_SVPWM_Handle_t SvpwmHandle;

PLX_PIDQ_Obj_t PiDqObj;
PLX_PIDQ_Handle_t PiDqHandle;

struct CONTROL_VARS ControlVars;

/**
 * Initializes control task.
 */
void InitControl(void){
	uint16_t pwmPeriod = (uint16_t)(SYSCLK_HZ / 2 / PWM_HZ / 2);

	InitPWM(pwmPeriod);
	InitADC();

	ControlVars.Fs = CONTROL_HZ;
	InitPuVars(ControlVars.Fs);

	SvpwmHandle = PLX_SVPWM_init(&SvpwmObj,sizeof(SvpwmObj));
	PLX_SVPWM_configure(SvpwmHandle, pwmPeriod, 2);

	PiDqHandle = PLX_PIDQ_init(&PiDqObj,sizeof(PiDqObj));

	DisableActuation();
	ResetControl();
}

/**
 * Called when controls is stopped, disable actuators here.
 */
void DisableActuation(void){
	// disable actuators
	EALLOW;
 	EPwm1Regs.TZFRC.bit.OST=1;
	EPwm2Regs.TZFRC.bit.OST=1;
	EPwm3Regs.TZFRC.bit.OST=1;
 	EDIS;
}

/**
 * Enable actuators.
 */
void EnableActuation(void){

	ResetControl();

#ifndef KEEP_POWERSTAGE_OFF
	// enable actuators
	EALLOW;
	EPwm1Regs.TZCLR.bit.OST=1;
	EPwm2Regs.TZCLR.bit.OST=1;
	EPwm3Regs.TZCLR.bit.OST=1;
	EDIS;
#endif
}

/**
 * Called when control starts, initialize filters, regulators here.
 */
void ResetControl(void){
	ControlVars.task1StepCtr = 0;
	ControlVars.task2StepCtr = 0;
	ControlVars.backgroundTaskNumber = 0;

	ControlVars.CmpU = 0;
	ControlVars.CmpV = 0;
	ControlVars.CmpW = 0;

	EPwm1Regs.CMPA.half.CMPA = ControlVars.CmpU;
	EPwm2Regs.CMPA.half.CMPA = ControlVars.CmpV;
	EPwm3Regs.CMPA.half.CMPA = ControlVars.CmpW;

	PLX_PIDQ_configure(PiDqHandle, PuVars.wrefTsScaling, Calib.KpD, Calib.KiD, Calib.KpQ, Calib.KiQ);
	PLX_PIDQ_reset(PiDqHandle);
}

/**
 * Fast control task
 */
void ControlTask1(){
	ControlVars.task1StepCtr++;

	// analog inputs
	SET_OPROBE(AdcOvrProbes.ADCRESULT0, AdcResult.ADCRESULT0);
	SET_OPROBE(AdcOvrProbes.ADCRESULT1, AdcResult.ADCRESULT1);
	SET_OPROBE(AdcOvrProbes.ADCRESULT2, AdcResult.ADCRESULT2);

	// process current measurements
	int16_t iu, iv;

	iu = AdcOvrProbes.ADCRESULT0 - AIn.IuOffset;
	iu = Q_MUL(iu, Calib.currentGain, Q_IV);
	iv = AdcOvrProbes.ADCRESULT1 - AIn.IvOffset;
	iv = Q_MUL(iv, Calib.currentGain, Q_IV);

	SET_OPROBE(AIn.Iu, iu);
	SET_OPROBE(AIn.Iv, iv);
	SET_OPROBE(AIn.Iw, -(iu + iv));

	// process DC voltage measurement
	int16_t vdc =  Q_MUL(AdcOvrProbes.ADCRESULT2, Calib.voltageGain, Q_IV);
	SET_OPROBE(AIn.Vdc, vdc);

	// other inputs
	SET_OPROBE(ControlVars.fluxPosition, 0);

	// set-points
	SET_OPROBE(ControlVars.IdSet, 0);
	SET_OPROBE(ControlVars.IqSet, 0);

	// calculate maximal achievable phase voltage based on DC-link voltage (max is 1/sqrt(3))
	int16_t vsmax = Q_MUL(AIn.Vdc, INV_SQRT3_Q15, 15);
	// apply margin for current regulator
	vsmax = Q_MUL(vsmax, (int16_t)(0.95 * 32768.0), 15);
	SET_OPROBE(ControlVars.Vsmax, vsmax);

	// clark transformation
	int16_t ia, ib;
	ia = AIn.Iu; // (2*AIn.Iu - AIn.Iv - AIn.Iw)/3
	ib = AIn.Iv - AIn.Iw;
	ib = Q_MUL(ib, INV_SQRT3_Q15, 15);

	SET_OPROBE(ControlVars.Ia, ia);
	SET_OPROBE(ControlVars.Ib, ib);

	// transform currents into rotating frame
	int16_t id, iq;
	PLX_FPMATH_lookupSinAndCos(ControlVars.fluxPosition, &ControlVars.fluxPosSin, &ControlVars.fluxPosCos);
	PLX_VECT_parkRot(ControlVars.Ia, ControlVars.Ib, &id, &iq, ControlVars.fluxPosSin, ControlVars.fluxPosCos);
	SET_OPROBE(ControlVars.Id, id);
	SET_OPROBE(ControlVars.Iq, iq);

	// determine errors
	SET_OPROBE(ControlVars.IdErr, ControlVars.IdSet - ControlVars.Id);
	SET_OPROBE(ControlVars.IqErr, ControlVars.IqSet - ControlVars.Iq);

	// update decoupling gains of current regulator (could happen at a lower rate, e.g. 1000 Hz)
	SET_OPROBE(ControlVars.we, 0);
	PLX_PIDQ_updateCrossCouplingGains(PiDqHandle,  ControlVars.we);
	PLX_PIDQ_applyCrossCouplingGains(PiDqHandle); // this function must not be interrupted!

	// calculate regulator output
	PLX_PIDQ_calculateOutput(PiDqHandle, ControlVars.IdErr, ControlVars.IqErr, &ControlVars.Vd, &ControlVars.Vq);

	// saturate output based on maximal allowable phase voltage
	int16_t vd = ControlVars.Vd;
	int16_t vq = ControlVars.Vq;
	PLX_VECT_satVector(ControlVars.Vsmax, &vd, &vq);

	// transform voltages back into stationary frame
	int16_t va, vb;
	PLX_VECT_invParkRot(vd, vq, &va, &vb, ControlVars.fluxPosSin, ControlVars.fluxPosCos);
	SET_OPROBE(ControlVars.Va, va);
	SET_OPROBE(ControlVars.Vb, vb);

	// determine SVPWM compare values
	int16_t svpwmLimitingGain;
	PLX_SVPWM_calcCompareValues(SvpwmHandle,
			ControlVars.Va, ControlVars.Vb, AIn.Vdc,
			&ControlVars.CmpU, &ControlVars.CmpV, &ControlVars.CmpW, &svpwmLimitingGain);

	// update ePWM compare values
	EPwm1Regs.CMPA.half.CMPA = ControlVars.CmpU;
	EPwm2Regs.CMPA.half.CMPA = ControlVars.CmpV;
	EPwm3Regs.CMPA.half.CMPA = ControlVars.CmpW;

	// apply SVPWM saturation to regulator outputs
	SET_OPROBE(ControlVars.svpwmLimitingGain, svpwmLimitingGain);
	ControlVars.VdSat = Q_MUL(vd, ControlVars.svpwmLimitingGain, 14);
	ControlVars.VqSat = Q_MUL(vq, ControlVars.svpwmLimitingGain, 14);

	// update regulator integrator based on saturated regulator outputs
	PLX_PIDQ_updateIntegrator(PiDqHandle, ControlVars.VdSat, ControlVars.VqSat);
}

/**
 * Slow control task
 */
void ControlTask2()
{
	ControlVars.task2StepCtr++;
	PLX_DELAY_US(300); // we add some fake "load" to demonstrate Task2 being interrupted by Task1
}

/**
 * Control taks run in background loop
 */
void ControlBackground(void)
{
	switch(ControlVars.backgroundTaskNumber)
	{
		case 0:
			PLX_DELAY_US(300);
			break;

		case 1:
			PLX_DELAY_US(1000);
			break;

		case 2:
			PLX_DELAY_US(600);
			break;

	}
	ControlVars.backgroundTaskNumber++;
	if(ControlVars.backgroundTaskNumber > 2)
	{
		ControlVars.backgroundTaskNumber = 0;
	}
}



