/*
   Copyright (c) 2014 by Plexim GmbH
   All rights reserved.

   A free license is granted to anyone to use this software for any legal
   non safety-critical purpose, including commercial applications, provided
   that:
   1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
   2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
 */

#include "includes.h"

// function prototypes
void DeviceInit(void);
void MemCopy(Uint16 *SourceAddr, Uint16 *SourceEndAddr, Uint16 *DestAddr);
void InitFlash();

// linker addresses, needed to copy code from flash to ram
extern Uint16 RamfuncsLoadStart, RamfuncsLoadEnd, RamfuncsRunStart;

// LED blink period/2
#define LED_BLINK_PERIOD_2 (500 * TICKS_PER_MS)

uint16_t LEDTimer;

struct CTRL_DISPATCH_VARS CtrlDispatchVars;

interrupt void ControlDispatcher(void)
{
	CtrlDispatchVars.interruptNesting++;

	PIL_beginInterruptCall();

#ifdef USE_ADC_INT
	AdcRegs.ADCINTFLGCLR.bit.ADCINT1 = 1;	// Clear ADCINT1 flag reinitialize for next SOC
	PieCtrlRegs.PIEACK.all = PIEACK_GROUP1; // Acknowledge interrupt to PIE
	IER |= M_INT1;
#else
	EPwm1Regs.ETCLR.bit.INT = 1;  // Clear INT flag for this timer
	PieCtrlRegs.PIEACK.all = PIEACK_GROUP3; // Acknowledge interrupt to PIE
	IER |= M_INT3;
#endif

	LEDTimer++;
	// re-enable interrupt to allow sub-task to be interrupted
	EINT;

	ControlTask1(); // this task is always dispatched

	bool_t dispatchTask2 = (CtrlDispatchVars.task2DispatchTimer == 0);

	CtrlDispatchVars.task2DispatchTimer++;
	if(CtrlDispatchVars.task2DispatchTimer >= TASK2_PERIOD)
	{
		CtrlDispatchVars.task2DispatchTimer = 0;
	}

	if(dispatchTask2)
	{
		CtrlDispatchVars.task2Active = TRUE;
		ControlTask2();
		CtrlDispatchVars.task2Active = FALSE;
	}

	DINT; // // TI expects interrupts are disabled before entering I$$REST
	CtrlDispatchVars.interruptNesting--;
}

void ResetControlDispatcher()
{
	CtrlDispatchVars.task2DispatchTimer = 0;
}

void InitControlDispatcher()
{
#ifdef USE_ADC_INT
	EALLOW;

	EPwm1Regs.ETSEL.bit.SOCAEN	= 1;		// Enable SOC on A group
	EPwm1Regs.ETSEL.bit.SOCASEL	= 2;		// Select SOC trigger on ePWM period
	EPwm1Regs.ETPS.bit.SOCAPRD 	= 1;		// Generate pulse on 1st event

	AdcRegs.INTSEL1N2.bit.INT1CONT  = 0;	// Disable ADCINT1 Continuous mode
	AdcRegs.INTSEL1N2.bit.INT1SEL 	= 2;    // setup EOC2 to trigger ADCINT1 (after Iu, Iv, Idc)
	AdcRegs.ADCCTL1.bit.INTPULSEPOS	= 1;	// ADCINT1 trips after AdcResults latch
	AdcRegs.INTSEL1N2.bit.INT1E     = 1;	// Enable ADCINT1
	EDIS;

	EALLOW;
	PieVectTable.ADCINT1 = &ControlDispatcher;
	EDIS;
	PieCtrlRegs.PIEIER1.bit.INTx1 = 1;
	IER |= M_INT1;
#else
	EPwm1Regs.ETSEL.bit.INTSEL = ET_CTR_PRD; // Select INT on period event
	EPwm1Regs.ETSEL.bit.INTEN = 1;           // Enable INT
	EPwm1Regs.ETPS.bit.INTPRD = ET_1ST; 	 // Generate interrupt on 1st event

	EALLOW;
	PieVectTable.EPWM1_INT = &ControlDispatcher;
	EDIS;
	PieCtrlRegs.PIEIER3.bit.INTx1 = 1;
	IER |= M_INT3;
#endif

	CtrlDispatchVars.interruptNesting = 0;
	CtrlDispatchVars.task2Active = FALSE;
	ResetControlDispatcher();
}

void BlinkLed(void)
{
	if(LEDTimer > LED_BLINK_PERIOD_2)
	{
		GpioDataRegs.GPBTOGGLE.bit.GPIO34 = 1;
		LEDTimer = 0;
	}
}

void main(void)
{
	DeviceInit();

	MemCopy(&RamfuncsLoadStart, &RamfuncsLoadEnd, &RamfuncsRunStart);

	InitFlash();

	EALLOW;
	// Configure IO for LED blinking
	GpioCtrlRegs.GPBMUX1.bit.GPIO34 = 0;
	GpioCtrlRegs.GPBDIR.bit.GPIO34 = 1;
	EDIS;

	DINT;

	IER = 0x0000;
	IFR = 0x0000;

	EALLOW;
	SysCtrlRegs.PCLKCR0.bit.TBCLKSYNC = 0;      // Stop all the TB clocks
	EDIS;

	InitCalibrations();
	InitControl();

	InitControlDispatcher();

	// configure serial line
	SCIInit(ONE_STOPBIT+ NO_PARITY + EIGHT_BITS, BAUD_RATE, LSPCLK_HZ);
	SCIWriteString("\n\rRS-232 Initialized.\n\r");

	// initialize PIL framework
	PIL_init();
	PIL_setLinkParams((unsigned char*)&PIL_D_Guid[0], (PIL_CommCallbackPtr_t)SCIPoll);
	PIL_setCtrlCallback((PIL_CtrlCallbackPtr_t)PilCallback);

	PilInitOverrideProbes();

	// Enable global Interrupts and higher priority real-time debug events:
	EINT;   // Enable Global interrupt INTM
	ERTM;   // Enable Global realtime interrupt DBGM

	EALLOW;
	SysCtrlRegs.PCLKCR0.bit.TBCLKSYNC = 1;         // Start all the timers synced
	EDIS;

	// get ready for PIL
	PIL_requestReadyMode();

	// main-loop
	for(;;) {
		BlinkLed();
		PIL_backgroundCall();
		ControlBackground();
	}
}
